import { NativeEventEmitter, NativeModules } from 'react-native';
const RCTAudioPlayer = NativeModules.AudioPlayer;
var PlayerSetupStatus;
(function (PlayerSetupStatus) {
    PlayerSetupStatus[PlayerSetupStatus["failed"] = -1] = "failed";
    PlayerSetupStatus[PlayerSetupStatus["unknown"] = 0] = "unknown";
    PlayerSetupStatus[PlayerSetupStatus["ready"] = 1] = "ready";
})(PlayerSetupStatus || (PlayerSetupStatus = {}));
;
var PlayerItemStatus;
(function (PlayerItemStatus) {
    PlayerItemStatus[PlayerItemStatus["buffering"] = 0] = "buffering";
    PlayerItemStatus[PlayerItemStatus["buffered"] = 1] = "buffered";
    PlayerItemStatus[PlayerItemStatus["playing"] = 2] = "playing";
    PlayerItemStatus[PlayerItemStatus["paused"] = 3] = "paused";
    PlayerItemStatus[PlayerItemStatus["waiting"] = 4] = "waiting";
})(PlayerItemStatus || (PlayerItemStatus = {}));
;
var PlayerInfo;
(function (PlayerInfo) {
    PlayerInfo["currentTime"] = "currentTime";
    PlayerInfo["duration"] = "duration";
    PlayerInfo["loadedTime"] = "loadedTime";
})(PlayerInfo || (PlayerInfo = {}));
;
var SupportedEvents;
(function (SupportedEvents) {
    SupportedEvents["playerStatus"] = "playerStatus";
    SupportedEvents["playerItemStatus"] = "playerItemStatus";
    SupportedEvents["playerInfo"] = "playerInfo";
})(SupportedEvents || (SupportedEvents = {}));
;
;
class AudioPlayer {
    constructor(url, statusHandler, eventEmitter = new NativeEventEmitter(RCTAudioPlayer)) {
        this.formatTime = (timeInseconds) => {
            const minutes = Math.floor(timeInseconds / 60);
            const seconds = Math.floor(timeInseconds % 60);
            return {
                seconds: timeInseconds,
                formatted: `${minutes}:${(seconds < 10 ? '0' : '') + seconds}`
            };
        };
        this.progressPercentage = (currentTimeInSeconds, duration) => {
            if (duration > 0) {
                const percent = (currentTimeInSeconds / duration) * 100;
                return percent > 100 ? 100 : percent;
            }
            return 0;
        };
        this.handlePlayerItemStatusChanges = (body) => {
            if (body === PlayerItemStatus.playing) {
                this.setStatus({ playing: true });
            }
            if (body === PlayerItemStatus.paused) {
                this.setStatus({ playing: false });
            }
        };
        this.handlePlayerInfo = (body) => {
            if (body.eventName === PlayerInfo.duration) {
                this.setStatus({ duration: this.formatTime(body.value) });
            }
            if (body.eventName === PlayerInfo.currentTime) {
                const currentTimeInSeconds = body.value;
                this.setStatus({
                    currentTime: this.formatTime(currentTimeInSeconds),
                    progress: this.progressPercentage(currentTimeInSeconds, this.status.duration.seconds)
                });
            }
            if (body.eventName === PlayerInfo.loadedTime) {
                this.setStatus({
                    percentLoaded: this.progressPercentage(body.value, this.status.duration.seconds)
                });
            }
        };
        this.toggleAudio = () => {
            if (this.status.playing) {
                RCTAudioPlayer.pause();
            }
            else {
                RCTAudioPlayer.play();
            }
        };
        this.seek = (timeInSeconds) => {
            RCTAudioPlayer.seek(timeInSeconds);
        };
        this.seekForwards = (timeInSeconds) => {
            RCTAudioPlayer.jump(timeInSeconds, false);
        };
        this.seekBackwards = (timeInSeconds) => {
            RCTAudioPlayer.jump(timeInSeconds, true);
        };
        this.url = url;
        this.statusHandler = statusHandler;
        this.eventEmitter = eventEmitter;
        this.status = {
            ready: false,
            playing: false,
            loading: false,
            duration: this.formatTime(0),
            currentTime: this.formatTime(0),
            progress: 0,
            percentLoaded: 0
        };
    }
    setStatus(changes) {
        this.status = Object.assign(Object.assign({}, this.status), changes);
        this.statusHandler(this.status);
    }
    create() {
        return new Promise((resolve, reject) => {
            const urlError = RCTAudioPlayer.prepare(this.url);
            if (urlError) {
                // problem with URL and player was never created
                this.setStatus({
                    loading: false
                });
                reject();
            }
            // set listener for all status changes on the asset
            this.eventEmitter.addListener(SupportedEvents.playerItemStatus, this.handlePlayerItemStatusChanges);
            // set listener for getting player info
            this.eventEmitter.addListener(SupportedEvents.playerInfo, this.handlePlayerInfo);
            // wait for player to be created
            this.eventEmitter.addListener(SupportedEvents.playerStatus, (body) => {
                if (body === PlayerSetupStatus.ready) {
                    this.setStatus({
                        ready: true,
                        loading: false
                    });
                    resolve();
                }
                if (body === PlayerSetupStatus.failed) {
                    this.setStatus({
                        loading: false
                    });
                    reject();
                }
            });
        });
    }
    destroy() {
        if (this.eventEmitter) {
            this.eventEmitter.removeAllListeners(SupportedEvents.playerStatus);
            this.eventEmitter.removeAllListeners(SupportedEvents.playerItemStatus);
            this.eventEmitter.removeAllListeners(SupportedEvents.playerInfo);
            RCTAudioPlayer.destroy();
        }
    }
}
export default AudioPlayer;
